/**
 * RESTful web-service client to retrieve data from 
 * the ADAPTER.DQ_TRANSMISSION_LOG database table for VAP.
 * 
 * Uses Spring RestTemplate, and inputs its web-service base-URL 
 * from server.properties.
 * 
 */

package gov.va.nvap.server.endpoint.dataquality;


import gov.va.nvap.common.validation.NullChecker;
import gov.va.nvap.service.audit.AuditException;

import gov.va.nvap.service.dataquality.DqTransmissionLogManager;
import gov.va.nvap.service.dataquality.DqTransmissionLogQueryRequest;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;


/**
 *
 * 
 */
public class DqTransmissionLogServiceEndpoint implements DqTransmissionLogManager {
    
    private static final Logger LOG = LoggerFactory.getLogger(DqTransmissionLogServiceEndpoint.class);
    
    // NOTE: The date format here matches the expected format for date-time on the service side
    static private final  String REQUEST_DATETIME_FORMAT = "yyyy-MM-dd.HH:mm:ss";
       
    private RestTemplate dqtlRestTemplate;
    
    private String serviceLocation;
    

    public DqTransmissionLogServiceEndpoint() {
        super();
    }    
    
    @Override
    public String getDqTransmissionLogData(final DqTransmissionLogQueryRequest logQueryRequest) throws AuditException {
        String response = null;
        
        // create the requestURL from the request parameters provided
        Map<String, String> vars = new HashMap<String, String>();
        StringBuilder sb = new StringBuilder();
        List<String> expressions = new ArrayList<String>();
        
        // get input parameters from the logQueryRequest parameter
        final Date runStartDate = logQueryRequest.getRunStartDate();
        final Date runEndDate = logQueryRequest.getRunEndDate();        
        final String filename = logQueryRequest.getFilename();
        final String eventType = logQueryRequest.getEventType();
        final String status = logQueryRequest.getStatus();
        
        // formats the Date parameters to specific string format
        final DateFormat dtFormatter = new SimpleDateFormat(REQUEST_DATETIME_FORMAT);
        
        if (!NullChecker.isNullOrEmpty(runStartDate)) {                     
            String runBeginStartDate = dtFormatter.format(runStartDate);
            expressions.add("runBeginStartDate={runBeginStartDate}");
            vars.put("runBeginStartDate", runBeginStartDate);
        } 
        
        if (!NullChecker.isNullOrEmpty(runEndDate)) { 
            String runFinishStartDate = dtFormatter.format(runEndDate);
            expressions.add("runFinishStartDate={runFinishStartDate}");
            vars.put("runFinishStartDate", runFinishStartDate);
        }
        
        if (!NullChecker.isNullOrEmpty(filename)) { 
            expressions.add("filename={filename}");
            vars.put("filename", filename);
        }
        
        if (!NullChecker.isNullOrEmpty(eventType) && !eventType.equals("ALL")) { 
            expressions.add("eventType={eventType}");
            vars.put("eventType", eventType);
        }   
        
        if (!NullChecker.isNullOrEmpty(status) && !status.equals("ALL")) { 
            expressions.add("status={status}");
            vars.put("status", status);
        }          
        
        // base URL set from .properties file
        sb.append(this.serviceLocation);
        sb.append("?");
        
        int length = expressions.size();
        for (int i = 0; i < length; i++) {
            if(0 == i) {
                sb.append(expressions.get(i));
            } else {
                sb.append("&");
                sb.append(expressions.get(i));
            }
        }        
        
        final String requestUrl = sb.toString();
      
        try {
            // make the actual REST request call using the RestTemplate
            if(LOG.isDebugEnabled()) {
                LOG.debug("calling DqTransmissionLogService to get log data with requestUrl pattern: "+requestUrl);
            }
            //System.out.println("calling DqTransmissionLogService to get log data with requestUrl pattern: "+requestUrl);
            for (Map.Entry<String, String> entry : vars.entrySet()) {
                if(LOG.isDebugEnabled()) {
                    LOG.debug("request template vars map: parameter name : " + entry.getKey() + " value : " + entry.getValue());
                }
                //System.out.println("request template vars map: parameter name : " + entry.getKey() + " value : " + entry.getValue());
            }
            ResponseEntity<String> entity = this.dqtlRestTemplate.getForEntity(requestUrl, String.class, vars);
            // get the response objects        
            HttpStatus responseHTTPStatusCode = entity.getStatusCode();
            int statusCode = responseHTTPStatusCode.value();
            MediaType responseContentType = entity.getHeaders().getContentType();
            String responseBodyStr = entity.getBody();
            // handle the response
            if(statusCode >= 200 & statusCode <= 299) {
                // success status
                if(responseContentType.equals(MediaType.APPLICATION_JSON)) {
                    // return the JSON string
                    response = responseBodyStr;
                } else {
                    // error in response                   
                    //LOG.error("DqTransmissionLogServiceEndpoint GET request received a 2xx status response with a non-JSON response body!: \n"+responseBodyStr);
                    throw new AuditException("DqTransmissionLogServiceEndpoint GET request received a 2xx status response with a non-JSON response body!: \n"+responseBodyStr);
                }
            } else {
                // error or something unexpected
                //LOG.error("DqTransmissionLogServiceEndpoint GET request received a non 2xx status code response!: \nHTTP status code: "+statusCode+" content-type: "+responseContentType.toString()+" response body: \n"+responseBodyStr);
                throw new AuditException("DqTransmissionLogServiceEndpoint GET request received a non 2xx status code response!: \nHTTP status code: "+statusCode+" content-type: "+responseContentType.toString()+" response body: \n"+responseBodyStr);
            }
        } catch (RestClientException e) {
            // request timed out or didn't connect
            //LOG.error("DqTransmissionLogServiceEndpoint GET request threw Exception: ",e);              
            throw new AuditException("DqTransmissionLogServiceEndpoint GET request threw RestClientException: "+e.getMessage(),e);
        }  
        if(LOG.isDebugEnabled()) {
            LOG.debug("returning DqTransmissionLogService response: "+response);
            //System.out.println("returning DqTransmissionLogService response: "+response);
        }
        // Uncomment this line and comment-out the following lines for the real data transfer
        return response;
        
        // file is in nvap-server/src/test/resources, and must be moved to src/main/resources, and would need to be allowed in by Maven filtering include for .json       
//        final String mockFilename = "sampleDQTransmissionLogServiceResponseJsonString.json";
//
//        InputStream inputStream = getClass().getClassLoader().getResourceAsStream(mockFilename);
//        String mockResponse = IOUtils.toString(inputStream, StandardCharsets.UTF_8.name());
//        return mockResponse;        
    }
    
    public void setDqtlRestTemplate(RestTemplate dqtlRestTemplate) {
        this.dqtlRestTemplate = dqtlRestTemplate;
    }

    public void setServiceLocation(String serviceLocation) {
        this.serviceLocation = serviceLocation;
    }
    
}
