package gov.va.nvap.web.app;

import gov.va.nvap.common.util.LogUtil;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.Properties;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import org.slf4j.LoggerFactory;
import org.slf4j.Logger;
import gov.va.nvap.propertylookup.PropertyLookup;
import gov.va.nvap.propertylookup.PropertyLookupLocal;
import gov.va.nvap.svc.consentmgmt.stub.dao.DescriptorDAO;
import gov.va.nvap.svc.consentmgmt.stub.data.Descriptor;
import java.util.HashMap;
import java.util.List;
import javax.ejb.EJB;
import javax.servlet.ServletContext;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;


/**
 * Web application life-cycle listener.
 *
 * @author David Vazquez Modified by: Asha Amritraj
 */
public class AppStateListener implements ServletContextListener {

    //static private Log LOG = LogFactory.getLog(AppStateListener.class);
    private static final Logger LOG = LoggerFactory.getLogger(AppStateListener.class);
    
    private PropertyLookup propertyLookup;  

    @Override
    public void contextDestroyed(final ServletContextEvent sce) {
        if (AppStateListener.LOG.isInfoEnabled()) {
            AppStateListener.LOG.info("******* VAP web application stopped. *******");            
        }
    }

    @Override
    public void contextInitialized(final ServletContextEvent sce) {
        // Get the version from the MANIFEST.MF file
        InputStream in = null;
        InputStream loadProps = null;
        try {
            in = sce.getServletContext().getResourceAsStream("META-INF/MANIFEST.MF");

            final String version = this.getValueFromManifest(in, "Build-VAP");
            // Put the build and version into the context
            sce.getServletContext().setAttribute("version", version);

            Properties properties = new Properties();
        
            loadProps = sce.getServletContext().getResourceAsStream("/WEB-INF/static.properties");
            properties.load(loadProps);
            // load from static.properties
            String systemVersion = properties.getProperty("system.version");
            sce.getServletContext().setAttribute("systemVersion", systemVersion);
            // load from web.properties
            sce.getServletContext().setAttribute("organizationName", propertyLookup.getProperty("va.ehx.organizationName"));            
            sce.getServletContext().setAttribute("scheduledExportMin", propertyLookup.getProperty("scheduled.export.min"));
            sce.getServletContext().setAttribute("userguideUrl", propertyLookup.getProperty("userguide.pdf.url"));
            
            //add the descriptors to the ServletContext (application/global) scope 
            this.addDescriptors(sce.getServletContext());
            
            if (AppStateListener.LOG.isInfoEnabled()) {
                AppStateListener.LOG.info("******* VAP web application (v." + LogUtil.cleanLogMessage(systemVersion) + ") starting. *******");
            }            
        } catch (IOException ex) {
            //Logger.getLogger(AppStateListener.class.getName()).log(Level.SEVERE, null, ex);
            LOG.error("",ex);            
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException e) {
                    //Logger.getLogger(AppStateListener.class.getName()).log(Level.SEVERE, null, e);
                    LOG.error("",e); 
                }
            }
            if(loadProps != null) {
                try {
                    loadProps.close();
                } catch(IOException e) {
                    //Logger.getLogger(AppStateListener.class.getName()).log(Level.SEVERE, null, e);
                    LOG.error("",e);
                }
            }
        }
    }

    /**
     * Get the value from the MANIFEST File
     */
    private String getValueFromManifest(final InputStream in,
        final String header) {
        String ret = null;
        final LineNumberReader reader = new LineNumberReader(
            new InputStreamReader(in));

        try {
            String line = reader.readLine();
            while ((ret == null) && (line != null)) {
                if (line.startsWith(header)) {
                    ret = line.substring(line.indexOf(":") + 1).trim();
                }
                line = reader.readLine();
            }
        } catch (final IOException ex) {
            //Logger.getLogger(AppStateListener.class.getName()).log(Level.WARNING, null, ex);
            LOG.error("",ex);
            ret = null;
        }

        return ret;
    }
    
    private void addDescriptors(ServletContext sc) {
        WebApplicationContext springContext = WebApplicationContextUtils.getWebApplicationContext(sc);
        DescriptorDAO dao = (DescriptorDAO) springContext.getBean("DescriptorDAO");
        
        List<Descriptor> dees = dao.getAll();
        
        HashMap<String, String> hm = new HashMap<String, String>();

        for(Descriptor d : dees) {
            hm.put(d.getName(), d.getText());
        }
        
        sc.setAttribute("descriptors", hm);
    }
    
    @EJB(beanInterface = PropertyLookupLocal.class, beanName = "WebPropertyFileLookup")
    public void setPropertyLookup(PropertyLookup propertyLookup)
    {
        this.propertyLookup = propertyLookup;
    }   
}